#
# Copyright (C) 2016 Simon Dobson
# 
# Licensed under the GNU General Public Licence v.2.0
#

import epyc

import numpy


class SummaryExperiment(epyc.ExperimentCombinator):
    '''An experiment combinator that takes an underlying experiment and 
    returns summary statistics for some of its results. This only really makes
    sense for experiments that return lists of results, such as those conducted
    using RepeatedExperiment.

    When run, a summary experiment summarises the experimental
    results, creating a new set of results that include the mean and
    variance for each result that the underyling experiments
    generated. (You can also select which results to summarise.) The
    raw results are discarded. The new results have the names of the
    raw results with suffices for mean, median, and variance".

    The summarisation obviously only works on result keys coming from the
    underlying experiments that are numeric. The default behaviour is to try to
    summarise all keys: you can restrict this by providing a list of keys to the
    constructor in the summarised_results keyword argument.

    The summary calculations only include those experimental runs that succeeded,
    that is that have their status set to True. Failed runs are ignored.'''

    # Additional metadata
    UNDERLYING_RESULTS = 'repetitions'                         #: Metadata relement for the number of repetitions performed
    UNDERLYING_SUCCESSFUL_RESULTS = 'successful_repetitions'   #: Metadata elements for the number of repetitions summarised
    UNDERLYING_EXCEPTIONS = 'underlying_exceptions'            #: Metadata elements for any exceptions raised

    # Prefix and suffix tags attached to summarised result and metadata values
    MEAN_SUFFIX = '_mean'              #: Suffix the mean of the underlking values
    MEDIAN_SUFFIX = '_median'          #: Suffix the median of the underlking values
    VARIANCE_SUFFIX = '_variance'      #: Suffix the variance of the underlking values
    
    
    def __init__( self, ex, summarised_results = None ):
        '''Create a summarised version of the given experiment. The given
        fields in the experimental results will be summarised, defaulting to all.
        If there are fields that can't be sujmmarised (because they're not
        numbers), remove them here. 

        :param ex: the underlying experiment
        :param summarised_results: list of result values to summarise (defaults to all)'''
        super(epyc.SummaryExperiment, self).__init__(ex)
        self._summarised_results = summarised_results

    def _mean( self, k ):
        '''Return the tag associated with the mean of k.'''
        return k + self.MEAN_SUFFIX

    def _median( self, k ):
        '''Return the tag associated with the median of k.'''
        return k + self.MEDIAN_SUFFIX

    def _variance( self, k ):
        '''Return the tag associated with the variance of k.'''
        return k + self.VARIANCE_SUFFIX
    
    def _summarise( self, results ):
        '''Private method to generate a summary result dict from a list of result dicts
        returned by do() on the repetitions of the underlying experiment.
        By default we generate order, mean and variance for each value recorded.

        We drop from the calculations any experiments whose completion status
        was False, indicating an error.

        Override this method to create different summary statistics.

        results: an array of result dicts
        returns: a dict of summary statistics'''
        if len(results) == 0:
            return dict()
        else:
            summary = dict()

            # work out the fields to summarise
            allKeys = results[0][epyc.Experiment.RESULTS].keys()
            ks = self._summarised_results
            if ks is None:
                # if we don't restrict, summarise all keys
                ks = allKeys
            else:
                # protect against a key that's not present
                ks = [ k for k in ks if k in allKeys ]
                
            # add the summary statistics
            for k in ks:
                # compute summaries for all fields we're interested in
                vs = [ res[epyc.Experiment.RESULTS][k] for res in results ]
                summary[self._mean(k)]     = numpy.mean(vs)
                summary[self._median(k)]   = numpy.median(vs)
                summary[self._variance(k)] = numpy.var(vs)
                    
            return summary   

    def do( self, params ):
        '''Perform the underlying experiment and summarise its results.
        Our results are the summary statistics extracted from the results of
        the instances of the underlying experiment that we performed.

        We drop from the calculations any experiments whose completion status
        was False, indicating an error. Our own completion status will be#
        true unless we had an error summarising a field (usually caused by trying
        to summarise non-numeric data).

        We record the exceptions generated by any experiment we summarise under
        the metadata key :attr:`SummaryExperiment.UNDERLYING_EXCEPTIONS`

        :param params: the parameters to the underlying experiment
        :returns: the summary statistics of the underlying results'''

        # perform the underlying experiment
        results = self.experiment().run()
        if not isinstance(results, list):
            results = [ results ]

        # extract only the successful runs
        sresults = [ res for res in results if res[epyc.Experiment.METADATA][epyc.Experiment.STATUS] ]
        exs = [ res[epyc.Experiment.METADATA][epyc.Experiment.EXCEPTION] for res in results if not res[epyc.Experiment.METADATA][epyc.Experiment.STATUS] ]
               
        # add extra values to out metadata record
        self._metadata[self.UNDERLYING_RESULTS]            = len(results)
        self._metadata[self.UNDERLYING_SUCCESSFUL_RESULTS] = len(sresults)
        self._metadata[self.UNDERLYING_EXCEPTIONS]          = exs
        
        # construct summary results
        return self._summarise(sresults)




# Simulation "lab" experiment management, sequential version
#
# Copyright (C) 2016 Simon Dobson
# 
# Licensed under the GNU General Public Licence v.2.0
#

import epyc

import collections


class Lab(object):
    '''A laboratory for computational experiments.

    A :class:`Lab` conducts an experiment at different points in a
    multi-dimensional parameter space.  The default performs all the
    experiments locally; sub-classes exist to perform remote parallel
    experiments.

    A :class:`Lab` stores its result in a notebook, an instance of :class:`LabNotebook`.
    By default the base :class:`Lab` class uses an in-memory notebook, essentially
    just a dict; sub-clasess use persistent notebooks to manage larger
    sets of experiments.
    '''

    def __init__( self, notebook = None ):
        '''Create an empty lab.
 
        :param notebook: the notebook used to store results (defaults to an empty :class:`LabNotebook`)'''
        if notebook is None:
            self._notebook = epyc.LabNotebook()
        else:
            self._notebook = notebook
        self._parameters = dict()

    def open( self ):
        '''Open a lab for business. Sub-classes might insist the they are
        opened and closed explicitly when experiments are being performed.
        The default does nothing.'''
        pass

    def close( self ):
        '''Shut down a lab. Sub-classes might insist the they are
        opened and closed explicitly when experiments are being performed.
        The default does nothing.'''
        pass

    def updateResults( self ):
        '''Update the lab's results. This method is called by all other methods
        that return results in some sense, and may be overridden to let the results
        "catch up" with external processing. The default does nothing.'''
        pass
    
    def addParameter( self, k, r ):
        '''Add a parameter to the experiment's parameter space. k is the
        parameter name, and r is its range.

        :param k: parameter name
        :param r: parameter range'''

        # if range is a single value, make it a list
        if isinstance(r, basestring) or not isinstance(r, collections.Iterable):
            r = [ r ]
        self._parameters[k] = r

    def parameters( self ):
        '''Return a list of parameter names.

        :returns: a list of parameter names'''
        return self._parameters.keys()

    def __len__( self ):
        '''The length of an experiment is the total number of data points
        that will be explored. 

        :returns: the length of the experiment'''
        n = 1
        for p in self.parameters():
            n = n * len(self._parameters[p])
        return n
        
    def __getitem__( self, k ):
        '''Access a parameter range using array notation.

        :param k: parameter name
        :returns: the parameter range'''
        return self._parameters[k]

    def __setitem__( self, k, r ):
        '''Add a parameter using array notation.

        :param k: the parameter name
        :param r: the parameter range'''
        return self.addParameter(k, r)

    def _crossProduct( self, ls ):
        '''Internal method to generate the cross product of all parameter
        values, creating the parameter space for the experiment.

        :param ls: an array of parameter names
        :returns: list of dicts'''
        p = ls[0]
        ds = []
        if len(ls) == 1:
            # last parameter, convert range to a dict
            for i in self._parameters[p]:
                dp = dict()
                dp[p] = i
                ds.append(dp)
        else:
            # for other parameters, create a dict combining each value in
            # the range to a copy of the dict of other parameters
            ps = self._crossProduct(ls[1:])
            for i in self._parameters[p]:
                for d in ps:
                    dp = d.copy()
                    dp[p] = i
                    ds.append(dp)

        # return the complete parameter space
        return ds
           
    def parameterSpace( self ):
        '''Return the parameter space of the experiment as a list of dicts,
        with each dict mapping each parameter name to a value.

        :returns: the parameter space as a list of dicts'''
        ps = self.parameters()
        if len(ps) == 0:
            return []
        else:
            return self._crossProduct(ps)
    
    def runExperiment( self, e ):
        '''Run an experiment over all the points in the parameter space.
        The results will be stored in the notebook.

        :param e: the experiment'''

        # create the parameter space
        ps = self.parameterSpace()

        # run the experiment at each point
        nb = self.notebook()
        for p in ps:
            # print "Running {p}".format(p = p)
            res = e.set(p).run()
            nb.addResult(res)

        # commit the results
        nb.commit()

    def notebook( self ):
        '''Return the notebook being used by this lab.

        :returns: the notebook'''
        return self._notebook
    
    def results( self ):
        '''Retrieve the list of :term:`results dict` hashes (each of which contains the
        point at which the experiment was evaluated to get this
        result).

        :returns: a list of results dicts'''
        self.updateResults()
        return self.notebook().results()

    def dataframe( self ):
        '''Return the results as a pandas DataFrame.

        :returns: the resulting dataset as a DataFrame'''
        self.updateResults()
        return self.notebook().dataframe()
    
    def ready( self ):
        '''Test whether all the results are ready, that is none are
        pending.

        :returns: True if the results are in'''
        self.updateResults()
        return (len(self.notebook().pendingResults()) == 0)





#
# Copyright (C) 2016 Simon Dobson
# 
# Licensed under the GNU General Public Licence v.2.0
#

import epyc

import numpy


class ExperimentCombinator(epyc.Experiment):
    '''An experiment that wraps-up another, underlying experiment. This is an abstract
    class that just provides the common wrapping logic.

    Experiment combinators aren't expected to have parameters of their own: they
    simply use the parameters of their underlying experiment.'''

    def __init__( self, ex ):
        '''Create a combinator based on the given experiment.

        ex: the underlying experiment'''
        super(epyc.ExperimentCombinator, self).__init__()
        self._experiment = ex

    def experiment( self ):
        '''Return the underlying experiment.

        returns: the underlying experiment'''
        return self._experiment

    def configure( self, params ):
        '''Configuring the summary configures the underlying experiment.

        params: the experiment's parameters'''
        self.experiment().configure(params)

    def deconfigure( self ):
        '''De-configuring the summary de-configures the underlying experiment.'''
        self.experiment().deconfigure()

    def parameters( self ):
        '''Return the current experimental parameters, taken from the
        underlying experiment.

        returns: the parameters,'''
        return self.experiment().parameters()



#!/usr/bin/env python

# Setup for epyc
#
# Copyright (C) 2016 Simon Dobson
# 
# Licensed under GPL 2.0
#

from setuptools import setup

with open('README.rst') as f:
    longDescription = f.read()

setup(name = 'epyc',
      version = '0.12.1',
      description = 'Python computational experiment management',
      long_description = longDescription,
      url = 'http://github.com/simoninireland/epyc',
      author = 'Simon Dobson',
      author_email = 'simon.dobson@computer.org',
      license = 'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
      classifiers = [ 'Development Status :: 4 - Beta',
                      'Intended Audience :: Science/Research',
                      'Intended Audience :: Developers',
                      'Programming Language :: Python :: 2.7',
                      'Topic :: Scientific/Engineering' ],
      packages = [ 'epyc' ],
      zip_safe = True,
      install_requires = [ "ipython", "pyzmq", "ipyparallel", "dill", "pandas", ])





# Experiment combinator to run the same experiment several times and
# return only the summary statistics
#
# Copyright (C) 2016 Simon Dobson
# 
# Licensed under the GNU General Public Licence v.2.0
#

import epyc


class RepeatedExperiment(epyc.ExperimentCombinator):
    '''A experiment combinator that takes a "base" experiment and runs it
    several times. This means you can define a single experiment separate
    from its repeating logic.

    When run, a repeated experiment runs a number of repetitions of the underlying
    experiment at the same point in the parameter space. The result of the
    repeated experiment is the list of results from the underlying experiment.
    If the underlying experiment itself returns a list of results, these are all
    flattened into a single list.'''

    def __init__( self, ex, N ):
        '''Create a repeated version of the given experiment.

        :param ex: the underlying experiment
        :param N: the number of repetitions to perform'''
        super(epyc.RepeatedExperiment, self).__init__(ex)
        self._N = N

    def repetitions( self ):
        '''Return the number of repetitions of the underlying experiment
        we expect to perform.

        returns: the number of repetitions'''
        return self._N

    def do( self, params ):
        '''Perform the number of repetitions we want.

        params: the parameters to the experiment
        returns: a list of result dicts'''
        N = self.repetitions()
        results = []
        for i in xrange(N):
            res = self.experiment().run()
            if isinstance(res, list):
                results.extend(res)
            else:
                results.append(res)
        return results

    def report( self, params, meta, res ):
        '''Return just the list of results, don't add any more metadata.

        :param params: the parameters we ran under
        :param meta: the metadata for this run
        :param res: the list of results
        :returns: the list of results'''
        return res
    
